//
// rotate.cc
//
// ASCII art rotation tool.
//
// This program reads in a text file specified on the command line,
// one that contains a rectagular grid of ASCII art, rotates it 90
// degrees clockwise, and then outputs the result into a similarly
// formatted text file.
//
// Usage: ./rotate <image file name>
//
#include <iostream>
#include <fstream>
#include <string>

struct asciiArt {
  char** image; // An array of arrays of characters.
  int width;    // Length of each row array.
  int height;   // Number of row arrays.
};

//
// readImage(fileName):
//
// Attempts to read an ASCII art image file with the specified
// `fileName`. The top line gives the rows and columns of the image, a
// `height` and a `width`. The subsequent `height` lines contain a
// sequence of characters, that row of the image.
//
// Returns a pointer to an `asciiArt` struct, or the null pointer if
// that file wasn't able to be opened.
//
asciiArt* readImage(std::string fileName) {
  // open the file for reading
  std::ifstream imageFile;
  imageFile.open(fileName);
  if (!imageFile.is_open()) return nullptr;

  // get the header line with height and width
  int width;
  int height;
  imageFile >> height;
  imageFile >> width;
  std::string cruft;
  std::getline(imageFile,cruft);

  // allocate the image rows array
  char** image = new char* [height];
  // get each image row line
  for (int r = 0; r < height; r++) {

    // Read a line from the file.
    std::string line;
    std::getline(imageFile,line,'\n');

    // Allocate that row of characters.
    image[r] = new char[width];
    for (int c=0; c<width; c++) {
      // fill in each character of that row
      if (c < line.length()) {
	image[r][c] = line[c];
      } else {
	image[r][c] = ' ';
      }
    }

  }

  // close the file and return a new image
  imageFile.close();  
  return new asciiArt {image,width,height};
}

//
// writeImage(fileName,art):
//
// Attempts to write an ASCII art image file with the specified
// `fileName` using the information stored in an `asciiArt` struct
// pointed at by `art`. The file format is the same as what's
// specified in `readImage` above.
//
// Returns `true` if successful, `false` if that file wasn't able to
// be opened.
//
bool writeImage(std::string fileName, asciiArt* art) {

  // open the file for writing
  std::ofstream imageFile;
  imageFile.open(fileName);
  if (!imageFile.is_open()) return false;

  // output the header with the number of rows and columns
  imageFile << art->height << " " << art->width << std::endl;

  // output each character of the image
  for (int r=0; r<art->height; r++) {
    // output one row of characters as a line
    for (int c=0; c<art->width; c++) {
      imageFile << art->image[r][c];
    }
    imageFile << std::endl;
  }

  // close the file
  imageFile.close();
  return true;
}

//
// outputArt = rotationOf(inputArt)
//
// This function takes a pointer to an `asciiArt` struct, and returns
// a pointer to a *new* `asciiArt` struct.
//
// The output image's characters should be the same as the input
// image's characters, only rotated 90 degrees clock-wise. That
// means the output will be an (n row) x (m column) image when the
// input image is an (m row) x (n column) image.
//
// You are to complete the code below for Exercise 2.  You need to
// allocate the array of row pointers of the output image, and then
// allocate each of the rows of characters for each line of the output
// image. Code that performs this kind of work can be found in the 
// code for `readImage` above.
//
// Note that this is why an `asciiArt` struct's `image` field is of type
// `char**`. The `image` field points to an array of `char*` pointers,
// each pointing at a row. In other words, it is an array of arrays
// with each element of type `char`,
//
// You access the `i`-th row with `outputArt->image[i]`. You can then
// acccess the `j`-th column in that row with `outputArt->image[i][j]`.
//
asciiArt* rotationOf(asciiArt* inputArt) {
  asciiArt* outputArt;

  // *** WRITE YOUR CODE BELOW HERE, REPLACING THIS ONE BOGUS LINE:
  //
  outputArt = inputArt; // NOPE.
  //
  // ***
  
  return outputArt;
}

//
// releaseImage(art):
//
// De-allocates the heap storage for the given `art`.
//
void releaseImage(asciiArt* art) {
  for (int i=0; i<art->height; i++) {
    delete [] art->image[i];
  }
  delete [] art->image;
  delete art;
}

//
// main(argc,argv):
//
// This program takes a name of a `.txt` file from the command line,
// (as `argv[1]`), reads it as an ASCII art image, rotates that to
// create a second image, and that writes that image out to a (new)
// text file.
//
// Example: The command line
//
//     ./rotate cheese1.txt
//
// will produce a file named `cheese1-90.txt`.
//
int main(int argc, char** argv) {

  // Look for the file name argument in `argv[1]`.
  if (argc < 2) {
    std::cout << "Error: missing file name!" << std::endl;
    std::cout << "Usage: " << argv[0] << " <image file name>.txt" << std::endl;
    return 0;
  }
  std::string artwork {argv[1]};

  // Parse the filename looking for the `.txt` extension.
  std::string base = artwork.substr(0,artwork.find_last_of('.'));
  std::string extn = artwork.substr(artwork.find_last_of('.'));  
  if (extn != ".txt") {
    std::cout << "Error: bad file extension '" << extn << "'!" << std::endl;
    std::cout << "Usage: " << argv[0] << " <image file name>.txt" << std::endl;
    return 0;
  }
  std::string artwork90 = base + "-90.txt";

  // Attempt to read the ASCII art file.
  asciiArt* art = readImage(artwork);
  if (art == nullptr) {
    std::cout << "Error: problem opening '" << artwork <<"' file!" << std::endl;
    return 0;
  }    

  /*** DELETE THESE LINES BELOW THAT PRINT THE ART. ***/
  for (int i=0; i<art->height; i++) {
    for (int j=0; j<art->width; j++) {
      std::cout << art->image[i][j];
    }
    std::cout << std::endl;
  }
  
  
  /*** UNCOMMENT THE LINES BELOW THAT ROTATE AND SAVE THE ART. ***
    
  // Create the rotated artwork, then write that to a (new) file.
  asciiArt* art90 = rotationOf(art);
  bool success = writeImage(artwork90,art90);
  if (!success) {
    std::cout << "Error: problem saving '" << (base+"-90.txt") <<"' file!" << std::endl;
    return 0;
  }

  // Release the heap memory for each image.
  releaseImage(art);    
  releaseImage(art90);    

  */
}    
